import os

# parse writeup content
lines = open('writeup.txt', 'r').readlines()
content = [line.strip() for line in lines if line.strip() != '']
name = content[0]
login = content[1]
content = content[2:]
features = []
while content:
    feature = content[0]
    implementation = content[1]
    argument = content[2]
    content = content[3:]
    inputs = []
    while content[0] != '=====':
        mesh = content[0]
        params = content[1].split(',')
        inputs.append([mesh, params])
        content = content[2:]
    features.append([feature, implementation, argument, inputs])
    content = content[1:]

# build Makefile targets from table of features
targets = [];
for feature in features:
    for input in feature[3]:
        base = input[0] + '_' + feature[0].strip().replace(' ', '_')
        targets.append([input[0], base + '_input', ''])
        base += '_%s'
        for parameter in input[1]:
            targets.append([input[0], \
                           base % parameter.strip().replace(' ', '_'), \
                           feature[2] % parameter])

# write Makefile
mf = open('Makefile', 'w')
if os.name == 'posix':
    print >> mf, 'IMGS= \\'
    for target in targets:
        print >> mf, 'output/' + target[1] + '.jpg \\'
    print >> mf, '\nall:\n\tcd src && $(MAKE)\n\t$(MAKE) imgs\n\nimgs: $(IMGS)\n'
    for target in targets:
        print >> mf, 'output/' + target[1] + '.jpg: input/' + target[0] + \
                     '.off\n\tsrc/meshpro $^ $(@:.jpg=.off) ' + target[2] + \
                     '\n\tsrc/meshview $(@:.jpg=.off) -output_image $@ -exit_immediately'
    print >> mf, '\nclean:\n\trm -f $(IMGS)\n\trm -f $(IMGS:.jpg=.off)\n\tcd src && $(MAKE) clean'
else:
    print >> mf, 'IMGS= \\'
    for target in targets:
        print >> mf, 'output\\' + target[1] + '.jpg \\'
    print >> mf, '\nall: $(IMGS)\n'
    for target in targets:
        print >> mf, 'output\\' + target[1] + '.jpg: input\\' + target[0] + \
                     '.off\n\tsrc\\meshpro $** $(@:.jpg=.off) ' + target[2] + \
                     '\n\tsrc\\meshview $(@:.jpg=.off) -output_image $@ -exit_immediately'
    print >> mf, '\nclean:\n\tdel $(IMGS) /q\n\tdel $(IMGS:.jpg=.off) /q'

# write writeup.html
page_header = \
"""<html>
<head>
  <title>COS 426 Assignment 2</title>
</head>
<body>
  <center><h2>COS 426 Assignment 2<br />%(name)s, %(login)s</h2></center>
  <hr /><br />
  <h2>Demonstration of Implemented Features</h2>"""
table_cell = \
"""      <td valign=top align=middle>
        <a href="output/%(file)s.jpg"><img width=300 src="output/%(file)s.jpg" /></a><br />
        %(caption)s
      </td>"""
wu =open('writeup.html', 'w')
print >> wu, page_header % {'name': name, 'login': login}
for feature in features:
    print >> wu, '  <h3>' + feature[0] + '</h3>\n  ' + feature[1]
    print >> wu, '  <br />\n<pre>src/meshpro [input] [output] ' + feature[2] + '</pre>'
    print >> wu, '  <table cellspacing=0 cellpadding=10 border=0>'
    for input in feature[3]:
        print >> wu, '    <tr>'
        base = input[0] + '_' + feature[0].strip().replace(' ', '_')
        print >> wu, table_cell % {'file': base + '_input', 'caption': 'Input'}
        base += '_%s'
        for parameter in input[1]:
            print >> wu, table_cell % {'file': base % parameter.strip().replace(' ', '_'), \
                                       'caption': parameter.strip()}
        print >> wu, '    </tr>'
    print >> wu, '  </table>'
print >> wu, \
"""  <hr />
  <h2>Feedback</h2>
  <dl>
    <dt>How long did you spend on this assignment?</dt>
    <dd>Your answer goes here</dd>

    <dt>Was it too hard, too easy, or just right?</dt>
    <dd>Your answer goes here</dd>

    <dt>What was the best part of the assignment?</dt>
    <dd>Your answer goes here</dd>

    <dt>What was the worst part of the assignment?</dt>
    <dd>Your answer goes here</dd>

    <dt>How could it be improved for next year?</dt>
    <dd>Your answer goes here</dd>

  </dl>
  <hr />
</body>
</html>"""
